/*
 * oled_ssd1306.c
 *
 * Created: 27.12.2019 10:41:08
 *  Author: GoldenAndy
 */ 

#include "oled_ssd1306.h"
#include <avr/pgmspace.h>
#include <string.h>

#define		SSDFLAG_DATS		(0b01000000)		//   
#define		SSDFLAG_DAT			(0b11000000)		//   
#define		SSDFLAG_COM			(0b10000000)		//   

#define SSD1306_INIT_LENGTH			25

#define SSD1306_ROWS				8
#define SSD1306_COLS				128


static uint8_t displayAddress = 0;

static struct {
	uint8_t swapX, swapY;
	} saveSwap = {0,0};

static struct {
	uint8_t leftBound;
	uint8_t rightBound;
	} bounds[SSD1306_ROWS];
static uint8_t videoram[SSD1306_ROWS][SSD1306_COLS];


const unsigned char PROGMEM initSequenceSSD1306[SSD1306_INIT_LENGTH]=
		{
			0xAE,		// off display
			0xA8,0x3F,	// multiplex ratio
			0xD3,0x00,	// Display offset
			0x40,		// Set start line
			0xA0,		// Set segment re-map  -  
			0xC0,		// Set COM output scan direction -   
			0xDA,0x12,	// Set COM pins hardware configuration
			0x81,0x7F,	//   () (000...0xFF)
			0xD5,0x80,	// set display clock divide ratio oscillator frequency
			0x8D,0x14,	// charge pump;	 0x8d,0x14 - Vcc by internal DC/DC, 0x8d,0x10 - Vcc external
			0xD9,0xF1,	// precharge period;	0xD9, 0xF1 - Vcc by internal DC/DC, 0xD9, 0x22 - Vcc external
			0xDB,0x20, // Set VCOMH deselect level
			0x20,0x00,  //  
			0xA4,		// Set entire display on/off
			0xA6,		// Sen norm/inverse display
			0xAF // AF - , AE -  
		};

uint8_t	ssd1306init(uint8_t displayAddr){
	displayAddress = displayAddr;
	uint8_t i,j=0;
	for (i=0; i<SSD1306_INIT_LENGTH; i++) {
		videoram[0][j++] = pgm_read_byte(initSequenceSSD1306 + i);
		videoram[0][j++] = SSDFLAG_COM;
	}
	uint8_t result = i2c_sendSSD1306data(displayAddr, SSDFLAG_COM, &videoram[0][0], SSD1306_INIT_LENGTH+SSD1306_INIT_LENGTH-1);
	ssd1306clearDisplay();
	if (result) result = ssd1306updateDisplay();
	return result;
}
	
uint8_t	ssd1306reInit(void){
	uint8_t result = ssd1306init(displayAddress);
	result &= ssd1306setOrientation(saveSwap.swapX, saveSwap.swapY);
	return result;
}
	
uint8_t ssd1306setOrientation(uint8_t swapX, uint8_t swapY){
	saveSwap.swapX = swapX;
	saveSwap.swapY = swapY;
	uint8_t cmd;
	if (swapX) cmd = 0xA1; else cmd = 0xA0;
	uint8_t result = i2c_sendSSD1306data(displayAddress, SSDFLAG_COM, &cmd, 1);
	if (swapY) cmd = 0xC8; else cmd = 0xC0;
	result &= i2c_sendSSD1306data(displayAddress, SSDFLAG_COM, &cmd, 1);
	return result;
}

uint8_t ssd1306setInversion(uint8_t isInvert){
	uint8_t cmd;
	if (isInvert) cmd = 0xA7; else cmd = 0xA6;
	return i2c_sendSSD1306data(displayAddress, SSDFLAG_COM, &cmd, 1);
}
		
uint8_t ssd1306on(void){
	uint8_t cmd = 0xAF;
	return i2c_sendSSD1306data(displayAddress, SSDFLAG_COM, &cmd, 1);
}

uint8_t ssd1306off(void){
	uint8_t cmd = 0xAE;
	return i2c_sendSSD1306data(displayAddress, SSDFLAG_COM, &cmd, 1);
}

uint8_t ssd1306setContrast(uint8_t contrastLevel){
	uint8_t cmd[3] = {0x81, SSDFLAG_COM, contrastLevel};
	return i2c_sendSSD1306data(displayAddress, SSDFLAG_COM, &cmd[0], 3);
}
	
	
/***********************    *****************************************/
//   .
//      
// ,         
	
uint8_t ssd1306updateDisplay(void){
	uint8_t	cmd[5] = { 0, SSDFLAG_COM, 0, SSDFLAG_COM, 0 };
	uint8_t result = 1, i;
	for (i=0; i<SSD1306_ROWS; i++) {
		if ( bounds[i].leftBound > bounds[i].rightBound ) continue;
		cmd[0] = 0x22;
		cmd[2] = i;
		cmd[4] = i;
		result &= i2c_sendSSD1306data(displayAddress, SSDFLAG_COM, &cmd[0], 5);
		cmd[0] = 0x21;
		cmd[2] = bounds[i].leftBound;
		cmd[4] = bounds[i].rightBound;
		result &= i2c_sendSSD1306data(displayAddress, SSDFLAG_COM, &cmd[0], 5);
		result &= i2c_sendSSD1306data(displayAddress, SSDFLAG_DATS, &videoram[i][bounds[i].leftBound], bounds[i].rightBound-bounds[i].leftBound+1);
		bounds[i].leftBound = 0xFF;
		bounds[i].rightBound = 0;
		if (!result) break;
	}
	return result;
}

	
void ssd1306clearDisplay(void){
	memset((void*)&videoram[0][0], 0x00, SSD1306_COLS*SSD1306_ROWS);
	for (uint8_t i=0; i<SSD1306_ROWS; i++) {
		bounds[i].leftBound = 0x00;
		bounds[i].rightBound = SSD1306_COLS-1;
	}
}
	
void ssd1306setPixel(uint8_t X, uint8_t Y, uint8_t color){
	//     X, Y
	// color == 0   , !=0 - 
	uint8_t xCol = X & 0x7F;
	uint8_t yRow = (Y & 0x38) >> 3;
	uint8_t yVal = 1 << (Y & 0x07);		//         
	if (color) videoram[yRow][xCol] |= yVal;
	else videoram[yRow][xCol] &= ~yVal;
	if ( bounds[yRow].leftBound  > xCol ) bounds[yRow].leftBound  = xCol;
	if ( bounds[yRow].rightBound < xCol ) bounds[yRow].rightBound = xCol;
}

void ssd1306pgmImage2videoRam(uint8_t* pgmImageArea){
	//    1024     
	uint16_t idx = 0;
	uint8_t* pgmArea = pgmImageArea;
	uint8_t* vRam = &videoram[0][0];
	while (idx < SSD1306_COLS * SSD1306_ROWS ) {
		vRam[idx++] = pgm_read_byte(pgmArea++);
	}
	for (uint8_t i=0; i<SSD1306_ROWS; i++) {
		bounds[i].leftBound = 0x00;
		bounds[i].rightBound = SSD1306_COLS-1;
	}

}
