/*
 * RDA5807M.c
 *
 * Created: 21.03.2019 16:36:12
 *  Author: GoldenAndy
 */ 

#include "RDA5807M.h"
#include <avr/pgmspace.h>
#include <string.h>

static rda5807mErrorType errorFlag = rda_ErrorTWI;
rda5807mStateType rda5807mState; // ,   .
 

const uint16_t RDA5807M_BandLowerLimits[] PROGMEM	= {8700,	7600,	7600,	6500 };
const uint16_t RDA5807M_BandHigherLimits[] PROGMEM	= {10800,	9100,	10800,	7600 };
const uint16_t RDA5807M_ChannelSpacingsDiv10[] PROGMEM = {100/10, 200/10, 50/10 /*, 25/5*/ };
const uint8_t RDA5807M_ChannelSpacingsMultiplier[] PROGMEM = {100/50, 200/50, 50/50 /*, 25/50*/ };
#define MINIMAL_SPACING_KHZ		50

//------------------------------------------- forward-
uint16_t makeFrequencyByChannel(uint16_t channel);
#ifndef RDA5807M_DISABLE_RDS
	void clearRds(void);
#endif

//-------------------------------------------    .
//   - clearError   
rda5807mErrorType rda5807mGetErrorState(uint8_t clearError){
	rda5807mErrorType err = errorFlag;
	if (clearError) errorFlag = rda_Ok;
	return err;
}

//-------------------------------------------   /  / 
uint16_t get5807register(uint8_t regAddr){
	uint16_t val;
	errorFlag = i2c_WriteAddrReadData( RDA5807M_I2C_ADDR_RANDOM_ACCESS, regAddr, &val, 2) ? rda_Ok : rda_ErrorTWI;
	val = ( ( (val >> 8) & 0xff) | ((val << 8) & 0xff00) );
	return val;
}

void set5807register(uint8_t regAddr, uint16_t val){
	val = ( ( (val >> 8) & 0xff) | ((val << 8) & 0xff00) );
	errorFlag = i2c_WriteAddrWriteData( RDA5807M_I2C_ADDR_RANDOM_ACCESS, regAddr, &val, 2) ? rda_Ok : rda_ErrorTWI;
}

void update5807register(uint8_t regAddr, uint16_t mask, uint16_t val) {
	set5807register(regAddr, (get5807register(regAddr) & (~mask) ) | val);
};

//-------------------------------------------  
rda5807mErrorType rda5807mInit(void) {
	//  id
	uint16_t iD = get5807register(RDA5807M_REG_ID);
	if (errorFlag != rda_Ok) return errorFlag;
	//   - ,  
	errorFlag = ( (iD & RDA5807M_VAL_IDMASK) == RDA5807M_VAL_ID ) ? rda_Ok : rda_ErrorBadID;
	if (errorFlag != rda_Ok) return errorFlag;
	
	set5807register(	RDA5807M_REG_CONFIG,
						RDA5807M_VAL_CLK_32KHZ						// - 32768 
						| RDA5807M_VAL_DHIZ | RDA5807M_VAL_DMUTE	// . mute  Hi-Z output
						| RDA5807M_VAL_SEEKUP  						//  
						| RDA5807M_VAL_RDSEN | RDA5807M_VAL_NEW		// RDS
						| RDA5807M_VAL_SKMODE						//   
						//| RDA5807M_VAL_ENABLE						// 
						); 
						
	rda5807mSetBand(RDA5807M_DEFAULT_BAND);
	rda5807mSetFreqStep(RDA5807M_DEFAULT_FREQ_STEP);
	rda5807mSetFrequency( pgm_read_word( &RDA5807M_BandLowerLimits[ rda5807mState.freqBand >> RDA5807M_BAND_SHIFT ]) );
	set5807register( RDA5807M_REG_CONFIG2, RDA5807M_VAL_REGCFG2_INITIAL );
	set5807register( RDA5807M_REG_VOLUME, RDA5807M_VAL_REGVOL_INITIAL );
	update5807register( RDA5807M_REG_BLEND, RDA5807M_VAL_EASTBAND65M, RDA5807M_VAL_EASTBAND65M);
	#ifndef RDA5807M_DISABLE_RDS
		clearRds();
	#endif
	return errorFlag;	
}

//-------------------------------------------  
void rda5807mSetBand(bandSelectType band){
	if (errorFlag != rda_Ok) return;
	update5807register(RDA5807M_REG_TUNING, RDA5807M_VAL_BAND_MASK, band);
	rda5807mState.freqBand = band;
}

//-------------------------------------------   
void rda5807mSetFreqStep(frequencyStepType step){
	if (errorFlag != rda_Ok) return;
	update5807register(RDA5807M_REG_TUNING, RDA5807M_VAL_FSTEP_MASK, step);
	rda5807mState.freqStep = step;
}

//-------------------------------------------   ,   

uint16_t rda5807mGetFrequency(void) {
	if (errorFlag != rda_Ok) return 0;
	// freq (kHz) =  (freqStep (kHz) * 10bit-CHAN_value) + band low freq(MHz)*1000
	uint16_t freq = get5807register(RDA5807M_REG_STATUS) & RDA5807M_READCHAN_MASK;//  10-bit CHAN_value
	return makeFrequencyByChannel(freq);
}

//-------------------------------------------   ,   
void rda5807mSetFrequency(uint16_t frequency) {
	if (errorFlag != rda_Ok) return;
	uint8_t bandIdx = rda5807mState.freqBand >> RDA5807M_BAND_SHIFT;
	uint16_t origin = pgm_read_word( &RDA5807M_BandLowerLimits[bandIdx] );
	if ( frequency < origin  ||
			frequency > pgm_read_word( &RDA5807M_BandHigherLimits[bandIdx] ) ) {
freqError:				
				errorFlag = rda_ErrorBadParam;
				return;
			}
	//       .
	frequency -= origin;
	//        (  10)     .
	// ,      - 50  -   10  ,    1/10  -  5.
	//  ,     100  200   -   2  4.
	uint16_t multiplier = frequency /  (MINIMAL_SPACING_KHZ/10);
	uint8_t spacingMul = pgm_read_byte( &RDA5807M_ChannelSpacingsMultiplier[rda5807mState.freqStep] );
	while (spacingMul>1) {
		multiplier >>= 1;
		spacingMul >>= 1;
	}
	//      ,        
	if ( frequency != multiplier  * pgm_read_byte( &RDA5807M_ChannelSpacingsDiv10[rda5807mState.freqStep]) ) goto freqError;
	// Attempt to tune to the requested frequency
	update5807register(	RDA5807M_REG_TUNING, 
						RDA5807M_VAL_CHAN_MASK | RDA5807M_VAL_TUNE,
						( multiplier << RDA5807M_CHAN_SHIFT) | RDA5807M_VAL_TUNE);
	#ifndef RDA5807M_DISABLE_RDS
		clearRds();
	#endif
}

//-------------------------------------------  
void rda5807mSetVolume(uint8_t volume){
	if (errorFlag != rda_Ok) return;
	if (volume > 15) {
		errorFlag = rda_ErrorBadParam;
		return;
	}
	update5807register( RDA5807M_REG_VOLUME, RDA5807M_VAL_VOLUME_MASK, volume );
} 

//-------------------------------------------   
void rda5807mSetBassBoost(bassBoostType boost){
	if (errorFlag != rda_Ok) return;
	update5807register(RDA5807M_REG_CONFIG, RDA5807M_VAL_BASS, boost ? RDA5807M_VAL_BASS : 0);
}

//-------------------------------------------  /
void rda5807mSetAudioMode(audioModeType monoStereo){
	if (errorFlag != rda_Ok) return;
	update5807register(RDA5807M_REG_CONFIG, RDA5807M_VAL_FMONO, monoStereo ? RDA5807M_VAL_FMONO : 0);
}

//-------------------------------------------   
void rda5807mSetMuteMode(muteModeType mute){
	if (errorFlag != rda_Ok) return;
	update5807register(RDA5807M_REG_CONFIG, RDA5807M_VAL_DMUTE, mute ? RDA5807M_VAL_DMUTE : 0);
}

//-------------------------------------------   
void rda5807mSetPowerMode(powerControlType power){
	if (errorFlag != rda_Ok) return;
	update5807register(RDA5807M_REG_CONFIG, RDA5807M_VAL_ENABLE, power ? RDA5807M_VAL_ENABLE : 0);
}

//-------------------------------------------  
void rda5807mShiftFrequency(tuningDirectionType tuneDir){
	if (errorFlag != rda_Ok) return;
	uint16_t freq = rda5807mGetFrequency();
	uint8_t step = pgm_read_byte( &RDA5807M_ChannelSpacingsDiv10[rda5807mState.freqStep] );
	uint8_t bandIdx = rda5807mState.freqBand >> RDA5807M_BAND_SHIFT;
	if (tuneDir == tuneUp) freq += step;
	else freq -= step;
	if (freq < pgm_read_word (&RDA5807M_BandLowerLimits[bandIdx]) || 
		freq > pgm_read_word (&RDA5807M_BandHigherLimits[bandIdx]) ) return;
	rda5807mSetFrequency(freq);
}

//-------------------------------------------  
void rda5807mSearch(tuningDirectionType searchDir){
	uint16_t regval = RDA5807M_VAL_SEEKSTART;
	if (searchDir == tuneUp ) regval |= RDA5807M_VAL_SEEKUP;
	update5807register(RDA5807M_REG_CONFIG, RDA5807M_VAL_SEEKUP | RDA5807M_VAL_SEEKSTART, regval);
	#ifndef RDA5807M_DISABLE_RDS
		clearRds();
	#endif
}


#ifndef RDA5807M_DISABLE_RDS
//------------------------------------------- RDS
//  ,  RDS   ....
	void getRDS(void) {
		static char segRDS[8];
		static char segRDS1[64];
		static uint8_t stationNameIdx = 0;
		uint16_t RDS[4];
		i2c_WriteAddrReadData( RDA5807M_I2C_ADDR_RANDOM_ACCESS, RDA5807M_REG_RDS1, &RDS, 8);
		uint8_t grp, n;
		uint16_t i;
		for (n=0; n<4; n++) RDS[n] = ( ( (RDS[n] >> 8) & 0xff) | ((RDS[n] << 8) & 0xff00) );
		grp = (RDS[1] >> 12) & 0xf;
		if (RDS[1] & 0x0800) {
			} else {
			switch (grp) {
				case 0:
					i = (RDS[1] & 3) << 1;
					segRDS[i] = (RDS[3] >> 8);
					segRDS[i + 1] = (RDS[3] & 0xFF);
					stationNameIdx = 0;
					rda5807mState.rdsStation[stationNameIdx] = 0;
					for (i = 0; i < 8; i++) {
						if (segRDS[i] > 31 && segRDS[i] < 127)
						rda5807mState.rdsStation[stationNameIdx++] = segRDS[i];
					}
					rda5807mState.rdsStation[stationNameIdx] = 0;
					break;
				case 2:
					i = (RDS[1] & 15) << 2;
					segRDS1[i] = (RDS[2] >> 8);
					segRDS1[i + 1] = (RDS[2] & 0xFF);
					segRDS1[i + 2] = (RDS[3] >> 8);
					segRDS1[i + 3] = (RDS[3] & 0xFF);
					for (i = 0; i < RDS_TEXT_LENGTH; i++) {
						rda5807mState.rdsText[i] = segRDS1[i];
					}
					rda5807mState.rdsText[RDS_TEXT_LENGTH] = 0;
					break;
				default:;
			}
		}
	}
#endif


//-------------------------------------------   
void rda5807mReadState(void){
	if (errorFlag != rda_Ok) return;
	uint8_t regs[4];
	errorFlag = i2c_WriteAddrReadData( RDA5807M_I2C_ADDR_RANDOM_ACCESS, RDA5807M_REG_STATUS, &regs, 4) ? rda_Ok : rda_ErrorTWI;
	uint16_t chan = regs[0];
	chan = chan*256 + regs[1];
	chan &= RDA5807M_READCHAN_MASK;
	rda5807mState.frequency		= makeFrequencyByChannel(chan);
	//rda5807mState.frequency		= rda5807mGetFrequency();
	rda5807mState.seekFail		= regs[0] & RDA5807M_HIVAL_SF;
	rda5807mState.seekComplete	= regs[0] & RDA5807M_HIVAL_STC;
	rda5807mState.stereo		= regs[0] & RDA5807M_HIVAL_ST;
	//   0x0A ,   ,    RDS
	// RDA5807M_HIVAL_RDSR		0x80	// RDS ready (1)
	// RDA5807M_HIVAL_RDSS		0x10	// RDS decoder synchronized (1)
	// RDA5807M_HIVAL_BLK_E		0x08	// (When RDS enabled) block E has been found (1)
	rda5807mState.strength	=(regs[2] & RDA5807M_HIVAL_RSSI_MASK ) >> 1;
	rda5807mState.isStation	= regs[2] & RDA5807M_HIVAL_FM_TRUE;
	rda5807mState.fmReady	= regs[3] & RDA5807M_LOVAL_FM_READY;
	//   0x0B ,   ,    RDS
	// RDA5807M_LOVAL_ABCD_E    0x10	// Data block E (1) or blocks A-D (0)
	// RDA5807M_LOVAL_BLERA     0x0C	// 2 bits error level in block A(RDS) or E(RBDS)
	// RDA5807M_LOVAL_BLERB     0x03	// 2 bits error level in block B(RDS) or E(RBDS)
	#ifndef RDA5807M_DISABLE_RDS
		static uint8_t noRdsCounter = 100;
		if ( regs[0] & RDA5807M_HIVAL_RDSR ) {
			noRdsCounter = 0;
			getRDS();
			rda5807mState.rdsReady	= 1;
			} else {
			noRdsCounter++;
			if (noRdsCounter > 64) {   //     RDS
				clearRds();
				rda5807mState.rdsReady	= 0;
				noRdsCounter = 0;
			}
		}
		
	#endif
}


//-------------------------------------------  -        

uint16_t makeFrequencyByChannel(uint16_t channel) {//  
	//     (  )     
	//       , ,        10.
	//     16      - ,   
	//    10   .
	uint16_t freq = ( channel  * pgm_read_byte( &RDA5807M_ChannelSpacingsDiv10[rda5807mState.freqStep]) ); //         .
	freq += pgm_read_word( &RDA5807M_BandLowerLimits[rda5807mState.freqBand >> RDA5807M_BAND_SHIFT]);	//   
	return freq;
}


//-------------------------------------------  -   RDS
#ifndef RDA5807M_DISABLE_RDS
	void clearRds(void){
		memset(&rda5807mState.rdsStation[0],0x00,sizeof(rda5807mState.rdsStation));
		memset(&rda5807mState.rdsText[0],0x00,sizeof(rda5807mState.rdsText));
	}
#endif

#undef MINIMAL_SPACING_KHZ